/*	===========================================================================
 UPC - EETAC - CSD - http://digsys.upc.edu

 * LAB4: Initial example for a free running FSM
 * Documentation location: https://digsys.upc.edu/csd/DEE/lab04/lab04.html
 * Hardware for prototyping: Arduino Uno board. Chip ATmega328P

Planning: 

- Step 1: Prepare the hardware, so that adapt the Arduino board to this 
particular application. For instance: add two switches and three LED circuits

- Step 2: Prepare the software. Hardware/software diagram. State diagram

    - Real-time measurement using millis() 
	
Version 2: FSM function containing both truth tables state_logic() 
and output_logic() in the same switch-case structure 
=============================================================================*/

/*	===========================================================================
	Include section
	=========================================================================*/

// External function libraries 


/*	===========================================================================
	Definitions
	======================================================================== */
const char pin_SW1 = 2; 
const char pin_SW2 = 3;

// LEDs for indicating the current state
const char pin_LED1 = 8; 
const char pin_LED2 = 9; 

// Common anode LED for generating the squared waveform (active-low)
const char pin_W_L = 10; 
 
// FSM list of states
const char STATE_0	= 'A'; 
const char STATE_1	= 'B'; 
const char STATE_2	= 'C'; 
const char STATE_3	= 'D'; 


// Timer period for generating a 12 Hz intermittency: 
// Waveform will be toggled after every TIMER_PERIOD
const unsigned long TIMER_PERIOD = 41; // 41.6666 ms   (24 Hz)


/*	===========================================================================
	Global RAM variables
	======================================================================== */

char	var_SW1	= 0; 
char	var_SW2	= 0; 
					
char	var_S = 0; // For representing the state in binary
char	var_Wave = 0; 

char	var_TW_flag = 0; // For timing half wave periods

char	var_current_state ;    // The internal memory of the FSM 

// For timing the square waveform: 
static unsigned long var_previous_time;   
//static unsigned long var_current_time;    // current time in ms
//static unsigned long var_time_elapsed;   //  time elapsed since last flag
 
/*	===========================================================================
	System initialisation
	=========================================================================*/
void setup() { 
// Configuring digital output pins
	pinMode(pin_LED1, OUTPUT);  
	pinMode(pin_LED2, OUTPUT);	
	pinMode(pin_W_L, OUTPUT);	

// Configuring digital input pins
	pinMode(pin_SW1, INPUT);  
	pinMode(pin_SW2, INPUT);  
	
// Serial monitor initialisation
	Serial.begin(9600); 
 
 //Initialise the FSM and the timer 
	var_current_state = STATE_0;	
	var_previous_time = 0; 	 

// Serial monitor ASCII message:
	Serial.println("Project LAB4 - DEE");
	Serial.println("Basic FSM architecture + digital I/O"); 
}

/*	===========================================================================
	Infinite loop (FSM)
	======================================================================== */
void loop() {       
	read_inputs(); 		// capture the switch voltage values into RAM variables
	fsm();
	write_outputs(); 	// Convert RAM variable values into pin output voltage	
}

/*	===========================================================================
	FSM function containing both truth tables state_logic() and output_logic() 
	in the same switch-case structure 
	======================================================================== */
void fsm(void){
	switch (var_current_state) {
	case STATE_0:
		// +++++++++++++++++++++ state_logic section ------------- 
		var_current_state = STATE_1; // jump immediately  

		// --------------------- output_logic section +++++++++++++ 
		var_S = 0;
		var_Wave = 0;

		break;
	
	case STATE_1:	
		// +++++++++++++++++++++ state_logic section ------------- 	
			// jump to the next state depending on input reading 
		if (var_SW1 == 1)	var_current_state = STATE_1;   
		else var_current_state = STATE_2; 

		// --------------------- output_logic section +++++++++++++ 
				// Here the square wave has to be generated
		update_time_flag();    //  to keep track of real time
		var_S = 1;
		if (var_TW_flag == 1){
			var_Wave = not(var_Wave); 	// Toggle waveform
			var_TW_flag = 0; 			// clear flag when used
		}		

		break;

	case STATE_2:	
		// +++++++++++++++++++++ state_logic section ------------- 		
		var_current_state = STATE_3;     
		
		// --------------------- output_logic section +++++++++++++ 
		var_S = 2;
		var_Wave = 0; 	// Here waveform output is OFF

		break;

	case STATE_3:
		// +++++++++++++++++++++ state_logic section ------------- 		
		if (var_SW2 == 1)	var_current_state = STATE_3;   
		else var_current_state = STATE_0; 
		
		// --------------------- output_logic section +++++++++++++ 
		var_S = 3;
		var_Wave = 1;	// Here waveform output is ON

		break;
	}	
}


/*	===========================================================================
	read_inputs 
	======================================================================== */
void read_inputs(void){
	if (digitalRead(pin_SW1) == 1) // Reading pins and organising variables
		var_SW1 = 1;
	else
		var_SW1 = 0;
   
	if (digitalRead(pin_SW2) == 1) // Reading pins and organising variables
		var_SW2 = 1;
	else
		var_SW2 = 0;
}

/*	===========================================================================
	write_outputs 
	======================================================================== */
void write_outputs(void){
// static char var_buf;  // local variable for saving partial results
// The output W_L (a single bit)
	if (var_Wave == 1)
		digitalWrite(pin_W_L, LOW); // Active-low connected LED 
	else
		digitalWrite(pin_W_L, HIGH);  

// This is a way to write output port pins using bitwise logic operations:
/*
	var_buf = var_S & 0b00000001; //Masking the bit of interest
   if (var_buf == 1)
       digitalWrite(pin_LED1, HIGH); // Active-high connected LED 
   else
       digitalWrite(pin_LED1, LOW);  
	
	var_buf = (var_S & 0b00000010)>>1;
   if (var_buf == 1)
       digitalWrite(pin_LED2, HIGH); // Active-high connected LED 
   else
       digitalWrite(pin_LED2, LOW);  
*/

// Another way to write to port pins, examine the differences, 
// which way is better?:
	switch (var_S) {
	case 0:
		digitalWrite(pin_LED2, LOW);
		digitalWrite(pin_LED1, LOW);		
		break;

	case 1: 
		digitalWrite(pin_LED2, LOW);
		digitalWrite(pin_LED1, HIGH);	
		break;

	case 2:
		digitalWrite(pin_LED2, HIGH);
		digitalWrite(pin_LED1, LOW);
		break;		
	
	case 3:
		digitalWrite(pin_LED2, HIGH);
		digitalWrite(pin_LED1, HIGH);
		break;
	} 

}



/*	===========================================================================
	Update time to calculate half periods for the output waveform
	======================================================================== */
void update_time_flag(void){
// local variables. "static allows as to follow them when debugging
static unsigned long var_current_time;    // current time in ms
static unsigned long var_time_elapsed;   //  time elapsed since last call

  // From internal CLK: 
	var_current_time = millis();
	var_time_elapsed =  var_current_time - var_previous_time;  

	if(var_time_elapsed  > TIMER_PERIOD) {
			// is the flag is set is because time elapsed is sampling period
		var_TW_flag = 1;   
			// to start another sampling period cycle
		var_previous_time = var_current_time; 
	}
}
